/*
 *  mod_musicindex.c
 *  mod_musicindex
 *
 *  $Id: mod_musicindex.c,v 1.50 2003/12/25 15:38:17 varenet Exp $
 *
 *  Created by Regis BOUDIN on Sat Dec 28 2002.
 *  Copyright (c) 2003 Regis BOUDIN
 *  Copyright (c) 2003 Thibaut VARENE
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation; either version 2.1, or (at your option)
 *  any later version.
 *
 */


/**
 * @mainpage
 * @section about About
 *	mod_musicindex is an Apache module aimed at being a C implementation
 *	of the Perl module <a href="http://search.cpan.org/dist/Apache-MP3/">
 *	Apache::MP3</a>. It allows nice displaying of directories containing
 *	MP3 or Ogg Vorbis files, including sorting them on various fields,
 *	streaming/downloading them, constructing playlist, and searching.
 *
 * @section features Features
 *	- Allow/deny file downloading.
 *	- Allow/deny file streaming.
 *	- Allow/deny file searching.
 *	- Default sorting order in Apache configuration and dynamic sorting while browsing.
 *	- Select fields to be shown in Apache configuration.
 *	- Custom folder picture (using cover.png or cover.jpg if found).
 *	- Cache subsystem (using flat files).
 *
 * @section dependencies Dependencies
 * 	- <a href="http://httpd.apache.org/download.cgi">Apache</a> (1.3 or 2).
 *	- <a href="http://www.vorbis.com/files/1.0/unix/libogg-1.0.tar.gz">libogg 1.0</a>.
 *	- <a href="http://www.vorbis.com/files/1.0/unix/libvorbis-1.0.tar.gz">libvorbis 1.0</a>.
 *	- <a href="ftp://ftp.mars.org/pub/mpeg/">libmad0</a>.
 *	- <a href="ftp://ftp.mars.org/pub/mpeg/">libid3tag0</a>.
 */

/**
 * @file 
 * Core file.
 *
 * This file is the core of the module. It contains Apache's mandatory stuff.
 *
 * @author Regis Boudin
 * @version $Revision: 1.50 $
 * @date 2003
 *
 * @warning Use it at your own risks!
 *
 * @todo Complete code documentation.
 * @todo custom playlist handling accross several directories.
 * @todo (whishlist) On the fly tag rewriting.
 * @todo (whishlist) On the fly reencoding.
 * @todo Considering the previous items, we might also need a handler for
 * streaming, otherwise we won't be able to restrain file downloading
 * (Apache::MP3 doesn't do anything better there fwiw).
 * @todo prepare the possibility to generate ices playlists.
 * @todo stabilize apache 2 code.
 * @todo i18n.
 * @todo enforce strict type policy.
 */

#include "mod_musicindex.h"
#include "playlist.h"
#include "config.h"
#include "html.h"
#include "inf.h"

static int handle_musicindex(request_rec *r)
{
	char *ifile;
	/* Get the module configuration */
	mu_config *conf;

#ifdef ENABLE_GETTEXT	/* ca va bien ici?? */
	setlocale(LC_ALL, "");
	bindtextdomain(PACKAGE, LOCALEDIR);
	textdomain(PACKAGE);
	bind_textdomain_codeset(PACKAGE, "UTF-8");
#endif	/* ENABLE_GETTEXT */

#ifdef __BUILD_FOR_APACHE2
	/* This test is necessary for apache 2 */
	if(strcmp(r->handler, DIR_MAGIC_TYPE))
		return DECLINED;
#endif
	conf = (mu_config *)ap_get_module_config(r->per_dir_config, &musicindex_module);

	r->allowed |= (1 << M_GET);
	/* if the request is not a GET and the module is not active, let the dir
	to another module */
	if ((r->method_number != M_GET) || (!(conf->options & MI_ACTIVE)))
		return DECLINED;

	/* This part mostly comes from mod_autoindex. If the requested dir does
	not end with a '/', generate a new request with it */
	if (r->uri[0] == '\0' || r->uri[strlen(r->uri) - 1] != '/') {
		if (r->args != NULL)
			ifile = ap_pstrcat(r->pool, ap_escape_uri(r->pool, r->uri),
						"/", "?", r->args, NULL);
		else
			ifile = ap_pstrcat(r->pool, ap_escape_uri(r->pool, r->uri),
						"/", NULL);

		ap_table_setn(r->headers_out, "Location", ap_construct_url(r->pool, ifile, r));
		return HTTP_MOVED_PERMANENTLY;
	}

	/* Depending on parameters from the HTTP request, call the adequate
	function to generate either a webpage or a m3u playlist */

	/* by default, we will generate a webpage */
	conf->options &= ~MI_PLAYLIST;
	switch (treat_args(r, conf)) {
		case 'A':
			return musicindex_list(r, conf);
		case 'L':
			return playlist_selected(r, conf);
		default:
			return 0;
	}

}

/**
 * Handler for requests on mp3 files.
 *
 * At the moment it can only forbid acces to files if neither streaming nor download is allowed.
 *
 * @param r Apache request_rec to get and send data
 *
 * @return Standard HTTP code (see httpd.h for more infos)
 */
int handle_mp3(request_rec *r)
{
	mu_config *conf;

#ifdef __BUILD_FOR_APACHE2
	/* This test is necessary for apache 2 */
	if(strcmp(r->handler,"audio/mpeg"))
		return DECLINED;
#endif
	conf = (mu_config *)ap_get_module_config(r->per_dir_config, &musicindex_module);

	if (!(conf->options&(MI_ALLOWSTREAM|MI_ALLOWDWNLD)))
		return HTTP_FORBIDDEN;
	
	if (r->args == NULL)
		return DECLINED;
	
	if (strcmp(r->args, "stream") == 0)
		return playlist_single(r, conf);
	
	return DECLINED;
}

/**
 * Handler for requests on ogg files.
 *
 * At the moment it can only forbid acces to files if neither streaming nor download is allowed.
 *
 * @param r Apache request_rec to get and send data
 *
 * @return Standard HTTP code (see httpd.h for more infos)
 */
int handle_ogg(request_rec *r)
{
	mu_config *conf;

#ifdef __BUILD_FOR_APACHE2
	/* This test is necessary for apache 2 */
	if((strcmp(r->handler,"application/ogg")) && (strcmp(r->handler,"audio/x-ogg")))
		return DECLINED;
#endif
	conf = (mu_config *)ap_get_module_config(r->per_dir_config, &musicindex_module);

	if (!(conf->options&(MI_ALLOWSTREAM|MI_ALLOWDWNLD)))
		return HTTP_FORBIDDEN;

	if (r->args == NULL)
		return DECLINED;
	
	if (strcmp(r->args, "stream") == 0)
		return playlist_single(r, conf);
	
	return DECLINED;
}

const command_rec musicindex_cmds[] = {
	AP_INIT_RAW_ARGS("MusicSortOrder", sort_order, NULL, OR_INDEXES,
		"can be : title album artist track disc length bitrate filetype genre"),
	AP_INIT_RAW_ARGS("MusicFields", set_fields, NULL, OR_INDEXES,
		"can be : title album artist track disc length bitrate filetype genre"),
	AP_INIT_NO_ARGS("MusicAllowStream", allow_stream, NULL, OR_INDEXES,
		"Enable streaming"),
	AP_INIT_NO_ARGS("MusicAllowDownload", allow_download, NULL, OR_INDEXES,
		"Enable download"),
	AP_INIT_NO_ARGS("MusicAllowSearch", allow_search, NULL, OR_INDEXES,
		"Enable searching (EXPERIMENTAL)"),
	AP_INIT_NO_ARGS("MusicLister", music_lister, NULL, OR_INDEXES,
		"Enable the Musicindex maker"),
	AP_INIT_TAKE1("MusicCachePath", set_cache_path, NULL, OR_INDEXES,
		"Set the cache absolute path, without trailing '/'."),
	AP_INIT_TAKE1("MusicPageTitle", set_page_title, NULL, OR_INDEXES,
		"Set the root title of the page."),
	{NULL}
};

#ifndef __BUILD_FOR_APACHE2 /* we build for apache 1.3 */

const handler_rec musicindex_handlers[] = {
	{DIR_MAGIC_TYPE, handle_musicindex},
	{"audio/mpeg", handle_mp3},
	{"audio/x-ogg", handle_ogg},
	{"application/ogg", handle_ogg},
	{NULL}
};

module MODULE_VAR_EXPORT musicindex_module = {
	STANDARD_MODULE_STUFF,
	NULL,				/**< initializer */
	create_musicindex_config,	/**< dir config creater */
	merge_musicindex_configs,	/**< dir merger */
	NULL,				/**< server config */
	NULL,				/**< merge server config */
	musicindex_cmds,		/**< command table */
	musicindex_handlers,		/**< handlers */
	NULL,				/**< filename translation */
	NULL,				/**< check_user_id */
	NULL,				/**< check auth */
	NULL,				/**< check access */
	NULL,				/**< type_checker */
	NULL,				/**< fixups */
	NULL,				/**< logger */
	NULL,				/**< header parser */
	NULL,				/**< child_init */
	NULL,				/**< child_exit */
	NULL				/**< post read-request */
};

#else /* we build for apache 2 */

static void register_hooks(apr_pool_t *p)
{
    ap_hook_handler(handle_musicindex,NULL,NULL,APR_HOOK_FIRST);
    ap_hook_handler(handle_mp3,NULL,NULL,APR_HOOK_FIRST);
    ap_hook_handler(handle_ogg,NULL,NULL,APR_HOOK_FIRST);
}

module AP_MODULE_DECLARE_DATA musicindex_module = {
    STANDARD20_MODULE_STUFF,
    create_musicindex_config,	/* create per-directory config structure */
    merge_musicindex_configs,	/* merge per-directory config structures */
    NULL,			/* create per-server config structure */
    NULL,			/* merge per-server config structures */
    musicindex_cmds,		/* command apr_table_t */
    register_hooks		/* register hooks */
};

#endif /* __BUILD_FOR_APACHE2 */
