/*
 *  playlist.c
 *  mod_musicindex
 *
 *  $Id: playlist.c,v 1.65 2003/12/25 15:38:17 varenet Exp $
 *
 *  Created by Thibaut VARENE on Thu Mar 20 2003.
 *  Copyright (c) 2003 Regis BOUDIN
 *  Copyright (c) 2003 Thibaut VARENE
 *   
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation; either version 2.1, or (at your option)
 *  any later version.
 *
 */

/**
 * @file 
 * Playlist management system.
 *
 * @author Regis Boudin
 * @author Thibaut Varene
 * @version $Revision: 1.65 $
 * @date 2003
 *
 * This is where the magic takes place. Here are the functions in charge of
 * analyzing the content of the folders requested by the client, and of
 * generating the structures and lists that will be used by the HTML
 * subsystem to generate the pretty output.
 *
 * @todo Find out if we trash "detect by content".
 */

#include "playlist.h"
#include "html.h"

/**
 * Creates a new musical entry (mp3, ogg or directory).
 *
 * @param pool Apache pool linked to the current request
 * @param head music entry structure which will follow this one in the list
 *
 * @return A new mu_ent
 */
mu_ent *new_ent(apr_pool_t *pool, mu_ent *head)
{
	mu_ent *p = (mu_ent *)ap_pcalloc(pool, sizeof(mu_ent));
	p->next = head;
	return p;
}

/**
 * MP3 content detection doesn't actually work.
 */
#undef DETECT_BY_CONTENT_NOT_BORKEN
#ifdef DETECT_BY_CONTENT_NOT_BORKEN
/* The following pieces of code are taken from xmms, in:
 *	Input/mpg123/common.c
 *	Input/mpg123/mpg123.c
 */
static unsigned long mpg123_convert_to_header(unsigned char *buf)
{
        return (buf[0] << 24) + (buf[1] << 16) + (buf[2] << 8) + buf[3];
}

/**
 * Checks for valid MP3 header.
 *
 * doesn't seem to work :P
 *
 * @param head The file header to check
 *
 * @return FALSE when the header is correct, TRUE otherwise.
 */
static short mpg123_head_check(unsigned long head)
{
        if ((head & 0xffe00000) != 0xffe00000)
                return FALSE;
        if (!((head >> 17) & 3))
                return FALSE;
        if (((head >> 12) & 0xf) == 0xf)
                return FALSE;
        if (!((head >> 12) & 0xf))
                return FALSE;
        if (((head >> 10) & 0x3) == 0x3)
                return FALSE;
        if (((head >> 19) & 1) == 1 && ((head >> 17) & 3) == 3 && ((head >> 16) & 1) == 1)
                return FALSE;
        if ((head & 0xffff0000) == 0xfffe0000)
                return FALSE;

        return TRUE;
}
#endif

/**
 * Reference table for digit coded genres.
 */
const char *mpg123_id3_genres[GENRE_MAX] = {
	"Blues", "Classic Rock", "Country", "Dance",
	"Disco", "Funk", "Grunge", "Hip-Hop",
	"Jazz", "Metal", "New Age", "Oldies",
	"Other", "Pop", "R&B", "Rap", "Reggae",
	"Rock", "Techno", "Industrial", "Alternative",
	"Ska", "Death Metal", "Pranks", "Soundtrack",
	"Euro-Techno", "Ambient", "Trip-Hop", "Vocal",
	"Jazz+Funk", "Fusion", "Trance", "Classical",
	"Instrumental", "Acid", "House", "Game",
	"Sound Clip", "Gospel", "Noise", "Alt",
	"Bass", "Soul", "Punk", "Space",
	"Meditative", "Instrumental Pop",
	"Instrumental Rock", "Ethnic", "Gothic",
	"Darkwave", "Techno-Industrial", "Electronic",
	"Pop-Folk", "Eurodance", "Dream",
	"Southern Rock", "Comedy", "Cult",
	"Gangsta Rap", "Top 40", "Christian Rap",
	"Pop/Funk", "Jungle", "Native American",
	"Cabaret", "New Wave", "Psychedelic", "Rave",
	"Showtunes", "Trailer", "Lo-Fi", "Tribal",
	"Acid Punk", "Acid Jazz", "Polka", "Retro",
	"Musical", "Rock & Roll", "Hard Rock", "Folk",
	"Folk/Rock", "National Folk", "Swing",
	"Fast-Fusion", "Bebob", "Latin", "Revival",
	"Celtic", "Bluegrass", "Avantgarde",
	"Gothic Rock", "Progressive Rock",
	"Psychedelic Rock", "Symphonic Rock", "Slow Rock",
	"Big Band", "Chorus", "Easy Listening",
	"Acoustic", "Humour", "Speech", "Chanson",
	"Opera", "Chamber Music", "Sonata", "Symphony",
	"Booty Bass", "Primus", "Porn Groove",
	"Satire", "Slow Jam", "Club", "Tango",
	"Samba", "Folklore", "Ballad", "Power Ballad",
	"Rhythmic Soul", "Freestyle", "Duet",
	"Punk Rock", "Drum Solo", "A Cappella",
	"Euro-House", "Dance Hall", "Goa",
	"Drum & Bass", "Club-House", "Hardcore",
	"Terror", "Indie", "BritPop", "Negerpunk",
	"Polsk Punk", "Beat", "Christian Gangsta Rap",
	"Heavy Metal", "Black Metal", "Crossover",
	"Contemporary Christian", "Christian Rock",
	"Merengue", "Salsa", "Thrash Metal",
	"Anime", "JPop", "Synthpop"
};

/**
 * Checks for valid MP3 filename extension.
 *
 * @param filename The filename to check.
 *
 * @return FALSE when the extension is correct, TRUE otherwise.
 */
static short mpg123_mp3_ext_check(char *filename)
{
	char *ext = strrchr(filename, '.');
	if (ext	&& (!strncasecmp(ext, ".mp2", 4) || !strncasecmp(ext, ".mp3", 4)))
		return FALSE;
	return TRUE;
}

/**
 * Converts an id3tag string in utf8.
 *
 * This function reads the id3tag from an MP3 file (if any) and converts
 * the result in the utf8 standard format.
 *
 * @warning Allocates memory. Don't forget to free().
 * 
 * @param tag The tag struct to parse
 * @param frameid The type of tag data searched
 * @param index The id3tag index value
 *
 * @return When possible, an utf8_t string, NULL otherwise.
 */
static id3_utf8_t *utf8_id3tag_findframe(struct id3_tag *tag, char const *frameid,
	unsigned short index)
{
	const struct id3_frame	*frame = NULL;
	const union id3_field	*field = NULL;
	const id3_ucs4_t	*ucs4 = NULL;

	if ((frame = id3_tag_findframe (tag, frameid, index))) {
		field = &frame->fields[1];
		if (id3_field_getnstrings(field)
			&& (ucs4 = id3_field_getstrings(field, index))) {
			return(id3_ucs4_utf8duplicate(ucs4));
			/* id3_ucs4_utf8duplicate() allocates memory.
			   Don't forget to free() after the function call. */
		}
	}
	return NULL;
}

/**
 * Fills in the information fields about OGG data.
 *
 * This function reads the ogg tags (using libvorbisfile) from the ogg file
 * and fills in the struct mu_ent fields accordingly. If the file is not an
 * ogg one, the function returns p = head.
 * 
 * @param pool Pool
 * @param head Head
 * @param in ogg file to parse
 * @param conf MusicIndex configuration paramaters struct
 * @param names Names
 * @param r Apache request_rec struct to handle log writings (debugging)
 *
 * @return When possible, struct mu_ent correctly set up
 */
mu_ent *make_ogg_entry(apr_pool_t *pool, mu_ent *head,
	FILE *in, mu_config *conf, mu_ent_names *names, request_rec *r)
{
	mu_ent			*p = head;
	OggVorbis_File		vf;
	vorbis_comment		*comment = NULL;
	char 			*t;
		
	if (!(ov_open(in, &vf, NULL, 0))) {
		p = new_ent(pool, head);
		p->filetype = FT_OGG;
	
		if ((comment = ov_comment(&vf, -1))) {
			if ((t = vorbis_comment_query(comment, "album", 0)))
				p->album = ap_pstrdup(pool, t);
			if ((t = vorbis_comment_query(comment, "artist", 0)))
				p->artist = ap_pstrdup(pool, t);
			if ((t = vorbis_comment_query(comment, "title", 0)))
				p->title = ap_pstrdup(pool, t);
			if ((t = vorbis_comment_query(comment, "tracknumber", 0)))
				p->track = atoi(t);
			if ((t = vorbis_comment_query(comment, "date", 0)))
				p->date = atoi(t);
			if ((t = vorbis_comment_query(comment, "tpos", 0)))	/* XXX does it exist? */
				p->posn = atoi(t);
			if ((t = vorbis_comment_query(comment, "genre", 0)))
				p->genre = ap_pstrdup(pool, t);
		}

		if (conf->options & (MI_QUICKPL)) {
			p->bitrate = p->length = 0;
		}
		else {
			p->bitrate = (long)ov_bitrate(&vf, -1);
			p->length = (long)ov_time_total(&vf, -1);
		}

		ov_clear(&vf);

	}

	return p;
}

/**
 * Fills in the information fields about MP3 data.
 *
 * This function reads the id3 tags (using libid3tag0) from the MP3 file
 * and fills in the struct mu_ent fields accordingly.
 * 
 * @param pool Pool
 * @param head Head
 * @param in MP3 file to parse
 * @param conf MusicIndex configuration paramaters struct
 * @param names Names
 * @param r Apache request_rec struct to handle log writings (debugging)
 *
 * @return When possible, struct mu_ent correctly set up
 *
 * @bug Reports wrong bitrate/length with VBR files.
 * @todo VBR handling (Xing headers decoding).
 */
mu_ent *make_mp3_entry(apr_pool_t *pool, mu_ent *head,
	FILE *in, mu_config *conf, mu_ent_names *names, request_rec *r) 
{
	mu_ent 			*p = head;

	struct id3_file		*id3struct = NULL;
	struct id3_tag		*tag = NULL;
	id3_utf8_t 		*utf8 = NULL;

	struct stat		filestat;
	
	struct mad_stream 	madstream;
	struct mad_frame	madframe;
	unsigned char		madinput_buffer[INPUT_BUFFER_SIZE];
	unsigned short		mp3gid = 0, isd = 0;

#ifdef DETECT_BY_CONTENT_NOT_BORKEN
	unsigned char		tmp[4];

	if (fread(tmp, 1, 4, in) != 4)
		goto exit;

	if (mpg123_head_check(mpg123_convert_to_header(tmp)))
#else
	if (mpg123_mp3_ext_check(names->filename))
#endif
		goto exit;
	
	mad_stream_init(&madstream);
	mad_frame_init(&madframe);
	
	fstat(fileno(in), &filestat);
	
	p = new_ent(pool, head);
	p->filetype = FT_MP3;
	p->size = filestat.st_size;
				
	if (conf->options & (MI_QUICKPL))
		p->bitrate = p->length = 0;
	else {
		/* First of all we try to find out the bitrate and length of
		 * the file */
		do {
			if (madstream.buffer == NULL || madstream.error == MAD_ERROR_BUFLEN) {
				size_t		madread_size, remaining;
				unsigned char	*madread_start;
				
				if (madstream.next_frame != NULL) {
					remaining = madstream.bufend-madstream.next_frame;
					memmove(madinput_buffer, madstream.next_frame, remaining);
					madread_start=madinput_buffer + remaining;
					madread_size = INPUT_BUFFER_SIZE - remaining;
				}
				else {
					madread_size = INPUT_BUFFER_SIZE;
					madread_start = madinput_buffer;
					remaining = 0;
				}
				
				madread_size = fread(madread_start, 1, madread_size, in);
				if (madread_size <= 0) {
					ap_log_rerror(APLOG_MARK, APLOG_ERR, r, "[musicindex] DBG: maderror madread_size <= 0 on %s", names->filename);
					goto exit;
					/* TODO traitement eventuel erreur lecture */
				}
	
				mad_stream_buffer(&madstream, madinput_buffer, madread_size + remaining);
				madstream.error = 0;
			}
	
	
				
			if (mad_frame_decode(&madframe,&madstream)) {
				if (MAD_RECOVERABLE(madstream.error)) {
					/* TODO traitement eventuel erreur recoverable */
					continue;
				}
				else {
					if (madstream.error == MAD_ERROR_BUFLEN)
						continue;
					else {
						ap_log_rerror(APLOG_MARK, APLOG_ERR, r, "[musicindex] DBG: maderror madstream.error unrecoverable on %s", names->filename);
						/* TODO traitement eventuel erreur unrecoverable */
						goto exit;
					}
				}
			}
		
			p->bitrate = madframe.header.bitrate * 1.024;
			p->length = filestat.st_size / (madframe.header.bitrate / 8);
			break;
		} while (1);
	}
	
	/* Then we check for an ID3 tag and read it if any */
	if ((id3struct = id3_file_open(names->filename, ID3_FILE_MODE_READONLY))) {
		if ((tag = id3_file_tag(id3struct)) && (tag->frames)) {
			if ((utf8 = utf8_id3tag_findframe(tag, ID3_FRAME_TITLE, 0))) {
				p->title = ap_pstrdup(pool, (char *)utf8);
				free(utf8);
			}

			if ((utf8 = utf8_id3tag_findframe(tag, ID3_FRAME_ARTIST, 0))) {
				p->artist = ap_pstrdup(pool, (char *)utf8);
				free(utf8);
			}
			
			if ((utf8 = utf8_id3tag_findframe(tag, ID3_FRAME_ALBUM, 0))) {
				p->album = ap_pstrdup(pool, (char *)utf8);
				free(utf8);
			}

			if ((utf8 = utf8_id3tag_findframe(tag, ID3_FRAME_YEAR, 0))) {
				p->date = atoi((char *)utf8);
				free(utf8);
			}

			if ((utf8 = utf8_id3tag_findframe(tag, ID3_FRAME_TRACK, 0))) {
				p->track = atoi((char *)utf8);
				free(utf8);
			}
			/* TPOS is Part Of a Set (aka disc number) */
			if ((utf8 = utf8_id3tag_findframe(tag, "TPOS", 0))) {
				p->posn = atoi((char *)utf8);
				free(utf8);
			}
			/* Some MP3s (especially VBRs) are nice enough to tag their length */
			if ((utf8 = utf8_id3tag_findframe(tag, "TLEN", 0))) {
				if ((atoi((char *)utf8) / 1000) > 0)
					p->length = atoi((char *)utf8) / 1000;
				free(utf8);
			}
			if ((utf8 = utf8_id3tag_findframe(tag, ID3_FRAME_GENRE, 0))) {
				if (*utf8 == '(') {	/* some encoders encaps id nb in brackets */
					(*utf8)++;
					mp3gid = atoi((char *)utf8);
					(*utf8)--;	/* MOUAHAHAHA MIAM MIAM */
					isd = 1;
				}
				else if ((isd = isdigit(*utf8)))
					mp3gid = atoi((char *)utf8);
				
				if (isd && (mp3gid < GENRE_MAX))
					p->genre = ap_pstrdup(pool, mpg123_id3_genres[mp3gid]);
				else
					p->genre = ap_pstrdup(pool, (char *)utf8);

				free(utf8);
			}

		}
		id3_file_close (id3struct);
	}

exit:
	return p;
}

/**
 * Add a file (and the content of the directory if requested) to the chain.
 *
 * This function creates a new entry from a file.
 * If the file is a directory and the recursive option is set, all its content
 * is also added.
 * If the file is an ogg vorbis or an mp3, it is simply added.
 * 
 * @param pool Pool
 * @param r Apache request_rec struct to handle log writings (debugging)
 * @param head Head
 * @param conf MusicIndex configuration paramaters struct
 * @param names Names
 *
 * @return When possible, struct mu_ent correctly set up
 */
mu_ent *make_music_entry(apr_pool_t *pool, request_rec *r,
	mu_ent *head, mu_config *conf, mu_ent_names *names)
{
	DIR 			*dir;
	struct dirent		*dstruct;
	mu_ent			*p = head;
	FILE			*in = NULL;
	unsigned short		i, checkresult = 0;
	char			*t, *fn, *uri;
#ifndef NO_CACHE
	mu_cache_data		*cachedata = NULL;
#endif
	
	if (!names) {
		names = ap_palloc(r->pool, sizeof(mu_ent_names));
		names->filename = ap_palloc(r->pool, MAX_STRING);
		names->uri = ap_palloc(r->pool, MAX_STRING);
		strcpy(names->filename, r->filename);
		strcpy(names->uri, r->parsed_uri.path);
	}

	t = names->uri + strlen(names->uri)-1;
	while (*t != '/')
		t--;
	t++;
	if (*t == '.')
		return p;

	if ((in = fopen(names->filename, "r"))) {
#ifndef NO_CACHE
		cachedata = ap_palloc(r->pool, sizeof(mu_cache_data));
		cachedata->name = names->filename;
#endif
		if ((ap_is_directory(pool, names->filename)) ) {
			fn = names->filename + strlen(names->filename);
			*fn++ = '/';
			*fn = '\0';
		
			uri = names->uri + strlen(names->uri) - 1;
			if (*uri++ != '/')
				*uri++ = '/';
			*uri = '\0';
				
			if (conf->options & MI_RECURSIVE) {
				conf->options &= (conf->play_recursive);

				if (!(dir = opendir(names->filename)))
					return head;
#ifndef NO_CACHE
				if ((conf->cache_path) && (cache_check_dir(r, conf, cachedata)))
					ap_log_rerror(APLOG_MARK, APLOG_ERR, r, "[musicindex] DBG: cache_check_dir failed");	/* XXX debug */
#endif
				while ((dstruct = readdir(dir))) {
					strcpy(fn, dstruct->d_name);
					strcpy(uri, dstruct->d_name);
					p = make_music_entry(pool, r, p, conf, names);	
				}
				closedir(dir);
				fclose(in);
				return p;
			}
			else if (!(conf->options & MI_PLAYLIST)) {
				p = new_ent(pool, head);
				p->filetype = FT_DIR;
			}
			fclose(in);
		}
#ifndef NO_CACHE
		/* from left to right: "is cache enabled?" && "is cache file ok to be read?" && "has cache file correctly been read?" */
		else if ((conf->cache_path) && !(checkresult = cache_check_file(r, conf, cachedata))
			&& ((p = cache_read_file(r, head, conf, cachedata)) != head))
				fclose(in);
#endif
		else if ((p = make_ogg_entry(pool, head, in, conf, names, r)) != head);
		/* ATTENTION! ov_open() takes complete possession of the file,
		   and ov_clear() closes it if ov_open() was successful.
		   We should NOT use fclose() in that case. */
		else if ((p = make_mp3_entry(pool, head, in, conf, names, r)) != head)
			fclose(in);
		else
			fclose(in);
	}

	if (p == head)
		return p;

	p->uri = ap_pstrdup(pool, names->uri);
	p->file = p->uri + strlen(r->parsed_uri.path);
	
	if (p->filetype == FT_DIR)
		return p;

	if (!(p->title)) {
#ifdef NO_TITLE_STRIP
		p->title = p->file;
#else
		/* Copy the name removing file extension and changing '_' to ' ' */
		p->title = ap_pstrndup(pool, p->file, strlen(p->file) - 4);
		for (i=0; p->title[i] != '\0'; i++)
			if (p->title[i] == '_')
				p->title[i] = ' ';
#endif	/* NO_TITLE_STRIP */
		/* We remove the trailing path if any. basename() ? */
		if (strrchr(p->title, '/')) {
			p->title = strrchr(p->title, '/');
			(p->title)++;
		}

	}
	
#ifndef NO_CACHE
	/* We put that here so that we do not create cache files for unhandled file types,
	 * and p->title is formated as we desire */
	if (checkresult == CA_CREATE)
		cache_write_file(r, p, conf, cachedata);
#endif
	
	if (!(conf->options & MI_SEARCH))
		return p;

	/* Here we go if we have an ongoing search request.
	 * We will return only the entries matching the search criteria */
	if (p->file && (ap_strcasestr(p->file, conf->search)))
		return p;
	if (p->artist && (ap_strcasestr(p->artist, conf->search)))
		return p;
	if (p->album && (ap_strcasestr(p->album, conf->search)))
		return p;
	if (p->title && (ap_strcasestr(p->title, conf->search)))
		return p;

	return head;
}

/**
 * Creates and sends a playlist or webpage with the content of a directory.
 *
 * This function generates a list containing all the files in the current
 * directory, either recursively or not. The list is then sent, either as a
 * m3u playlist or a web page.
 * 
 * @param r Apache request_rec struct
 * @param conf MusicIndex configuration paramaters struct
 *
 * @return 0 on success, error code otherwise.
 */
short musicindex_list(request_rec *r, mu_config *conf)
{
#ifdef __BUILD_FOR_APACHE2
	apr_dir_t	*dir;
#else
	DIR		*dir;
#endif
	mu_ent		*head = NULL;

#ifdef __BUILD_FOR_APACHE2
	if (apr_dir_open(&dir, r->filename, r->pool) != APR_SUCCESS) {
#else
	if (!(dir = ap_popendir(r->pool, r->filename))) {
#endif
		ap_log_rerror(APLOG_MARK, APLOG_ERR, r,
			"Can't open directory for music index: %s",
			r->filename);
		return HTTP_FORBIDDEN;
	}

	ap_pclosedir(r->pool, dir);

	if (conf->options & MI_PLAYLIST) {
		r->content_type = "audio/mpegurl";
		ap_table_setn(r->headers_out, "Content-Disposition",
			"filename = \"playlist.m3u\"");
	}
	else
		r->content_type = "text/html; charset=UTF-8";
	
	ap_send_http_header(r);

	if (r->header_only)
		return 0;
	
	ap_hard_timeout("send music list", r);

	head = make_music_entry(r->pool, r, NULL, conf, NULL);
	head = quicksort(head, NULL, conf);

	if (conf->options & MI_PLAYLIST)
		send_playlist(r, head, conf);
	else {
		send_head(r, conf);
		if (!(conf->options & MI_SEARCH))
			send_directories(r, head, conf);
		send_tracks(r, head, conf);
		send_foot(r, conf);
	}

	ap_kill_timeout(r);
	return 0;
}

/**
 * Generates and sends a playlist with only selected items.
 *
 * This function parses the requests, which is the form :
 * http://...../?file=file1&file=file2&file=file3...
 *
 * Note that in FORM requests, arguments have ' ' replaced by '+', other
 * "special" characters are changed to their URI equivalent. Here is the trick :
 * ap_unescape_url doesn't change the '+' to its URI equivalent. So we first
 * change ' ' to '+' and then call the apache function. Easy, isn't it ?
 * 
 * @param r Apache request_rec struct
 * @param conf MusicIndex configuration paramaters struct
 *
 * @return 0
 */
short playlist_selected(request_rec *r, mu_config *conf)
{
	const char		*args = r->args;
	char			*p;
	unsigned short		i;

	char			*fn, *uri;
	mu_ent_names		*names;
	mu_ent			*head = NULL;

	r->content_type = "audio/mpegurl";
	ap_table_setn(r->headers_out, "Content-Disposition",
		"filename = \"playlist.m3u\"");

	ap_send_http_header(r);

	if (r->header_only)
		return 0;
	ap_hard_timeout("send playlist", r);

	if ((args == NULL) || (*args == '\0')) {
		ap_kill_timeout(r);
		return 0;
	}

	names = ap_palloc(r->pool, sizeof(mu_ent_names));
	names->filename = ap_palloc(r->pool, MAX_STRING);
	names->uri = ap_palloc(r->pool, MAX_STRING);
	strcpy(names->filename, r->filename);
	strcpy(names->uri, r->parsed_uri.path);

	fn = names->filename + strlen(names->filename);
	*fn++ = '/';
	*fn = '\0';

	uri = names->uri + strlen(names->uri) - 1;
	if (*uri++ != '/')
		*uri++ = '/';
	*uri = '\0';

	while (args[0]) {
		p = ap_getword(r->pool, &args, '&');
		if (!strncmp(p, "file=", 5))
		{
			p += 5;
			strcpy(fn, p);
			for (i=0; fn[i]; i++) {
				if (fn[i] == '+')
					fn[i] = ' ';
			}
			ap_unescape_url(fn);
			strcpy(uri, fn);
			head = make_music_entry(r->pool, r, head, conf, names);
		}
	}

	head = quicksort(head, NULL, conf);
	send_playlist(r, head, conf);

	ap_kill_timeout(r);
	return 0;
}

/**
 * Generates a playlist for a single file
 *
 * This function takes the requested filename from the request and sends
 * the http link in a "audio/mpegurl" (aka playlist) file.
 * 
 * @param r Apache request_rec struct
 * @param conf MusicIndex configuration paramaters struct
 *
 * @return 0
 */
short playlist_single(request_rec *r, mu_config *conf)
{
	mu_ent_names		names;
	mu_ent			*head = NULL;

	r->content_type = "audio/mpegurl";
	ap_table_setn(r->headers_out, "Content-Disposition",
		"filename = \"playlist.m3u\"");

	ap_send_http_header(r);

	if (r->header_only)
		return 0;
	
	ap_hard_timeout("send playlist", r);

	names.filename = r->filename;
	names.uri = r->parsed_uri.path;
	
	head = make_music_entry(r->pool, r, head, conf, &names);
	send_playlist(r, head, conf);
	
	ap_kill_timeout(r);
	return 0;
}
